package vehicle;

import a_program.Program;
import a_program.VrPerson;
import com.jme3.asset.AssetManager;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.FastMath;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Node;
import com.jme3.scene.Spatial;
import com.jme3.scene.shape.Box;
import com.jme3.util.TangentBinormalGenerator;
import e_tco.Action3dElement;
import e_tco.GeometryAction;
import vehicle.Coupling.CouplingState;

/**
 * Air connections
 *
 * @author sindre
 */
public class AirCoupling implements Action3dElement {

    private AssetManager assetManager;
    private Vehicle vehicle;
    private int location; // Where the coupling is located, 1 = front, 2 = back
    private Vector3f target; // Target location for the inverse kinematics
    private Spatial heldBy;
    private CouplingState state = CouplingState.Loose;
    private AirCoupling connectedTo;

    private boolean open;   //  Whether the air valve is open or not

    //  Nodes
    private Node basePivot = new Node();

    private Node link1Pivot = new Node();
    private Node link2Pivot = new Node();
    private Node link3Pivot = new Node();

    private Node endPivot = new Node();
    private Node connectPivot = new Node();

    private Node airValvePivot = new Node();
    private AirValve airValve;

    private InverseKinematics IK = new InverseKinematics(
            new Node[]{basePivot, link1Pivot, link2Pivot, link3Pivot, connectPivot}, false);

    //  Offsets
    private float startToEndLength;

    public AirCoupling(AssetManager assetManager, Vehicle vehicle, int location) {
        this.assetManager = assetManager;
        this.vehicle = vehicle;
        this.location = location;

        Material matDebug = new Material(assetManager, "Common/MatDefs/Misc/ShowNormals.j3md");
        Material matTube = new Material(assetManager, "Common/MatDefs/Misc/Unshaded.j3md");
        Material matEnd = new Material(assetManager, "Common/MatDefs/Misc/Unshaded.j3md");
        matTube.setColor("Color", ColorRGBA.Orange);
        matEnd.setColor("Color", ColorRGBA.Gray);

        Box cube1Mesh = new Box(0.05f, 0.05f, 0.05f);
        GeometryAction kobbelGeo = new GeometryAction("AirCoupling", cube1Mesh, this);
        kobbelGeo.setMaterial(matDebug);

        TangentBinormalGenerator.setToleranceAngle(179);   // Stops some error warnings, models are funtioning

        airValve = new AirValve(assetManager, airValvePivot, this);
        basePivot.attachChild(airValvePivot);

        float l = -.09f;

        Box linkMesh = new Box(0.03f, l, 0.03f);
        Geometry link1Geo = new Geometry("Link1", linkMesh);
        link1Geo.setMaterial(matTube);
        basePivot.attachChild(link1Pivot);
        link1Pivot.attachChild(link1Geo);
        link1Pivot.setLocalTranslation(0, l, 0);

        Geometry link2Geo = new Geometry("Link2", linkMesh);
        link2Geo.setMaterial(matTube);
        link1Pivot.attachChild(link2Pivot);
        link2Pivot.attachChild(link2Geo);
        link2Pivot.setLocalTranslation(0, l * 2, 0);

        Geometry link3Geo = new Geometry("Link3", linkMesh);
        link3Geo.setMaterial(matEnd);
        link2Pivot.attachChild(link3Pivot);
        link3Pivot.attachChild(link3Geo);
        link3Pivot.setLocalTranslation(0, l * 2, 0);

        link3Pivot.attachChild(endPivot);
        endPivot.attachChild(connectPivot);
        endPivot.setLocalTranslation(0, l, 0);

        endPivot.attachChild(kobbelGeo);            //  For debugging

        target = endPivot.getWorldTranslation();
        startToEndLength = basePivot.getWorldTranslation().distance(target);
    }

    //  Update function, updates every frame
    public void tick(float tpf) {
        switch (state) {
            case Loose:
                target = basePivot.getWorldTranslation().add(new Vector3f(0, -startToEndLength * 1.5f, 0));
                inverseKinematics(target, tpf / 10f);
                break;
            case Held:
                inverseKinematics(target, tpf);
                break;
            case Connected:
                if (connectedTo != null) {
                    target = connectedTo.getConnectPivot().getWorldTranslation();
                    inverseKinematics(target, tpf);
                }
                break;
        }
    }

    public Node getBasePivot() {
        return basePivot;
    }

    public Node getEndPivot() {
        return endPivot;
    }

    public Node getConnectPivot() {
        return connectPivot;
    }

    public Node getAirValvePivot() {
        return airValvePivot;
    }

    public AirCoupling getConnectedTo() {
        return connectedTo;
    }

    public Vehicle getVehicle() {
        return vehicle;
    }

    public void setConnectedTo(AirCoupling c) {
        state = CouplingState.Connected;
        connectedTo = c;
    }

    public void connectTo(AirCoupling coupling) {
        state = CouplingState.Connected;
        connectedTo = coupling;
        coupling.setConnectedTo(this);
        heldBy = null;
        Program.updateCouplings();
    }

    public void disconnect() {
        state = CouplingState.Loose;
        connectedTo.setConnectedTo(null);
        connectedTo.setState(CouplingState.Loose);
        connectedTo = null;
        Program.updateCouplings();
    }

    public void setTarget(Vector3f target) {
        this.target = target;
    }

    public void hold(Spatial heldBy) {
        setState(CouplingState.Held);
        this.heldBy = heldBy;
        setTarget(heldBy.getWorldTranslation());
    }

    public void release() {
        heldBy = null;
        setState(CouplingState.Loose);
    }

    public Spatial heldBy() {
        return heldBy;
    }

    public CouplingState getState() {
        return state;
    }

    public void setState(CouplingState s) {
        state = s;
    }

    public void inverseKinematics(Vector3f target, float tpf) {
        IK.setActive(true);
        IK.setTarget(target);
        IK.controlUpdate(tpf);
    }

    public void setOpen(boolean open) {
        this.open = open;
    }

    public boolean getOpen() {
        return open;
    }

    private void attemptConnect() {
        VrPerson.airConnect(this);
    }

    public void airValveInteract() {
        airValve.click();
    }

    @Override
    public void click() {
        if (state != CouplingState.Connected) {
            attemptConnect();
        } else {
            disconnect();
        }
    }
}

//--------------------------------------
//  Air Valve
//--------------------------------------
class AirValve implements Action3dElement {

    boolean open = false;
    AssetManager assetManager;
    Node airValve;
    AirCoupling air;

    public AirValve(AssetManager assetManager, Node airValve, AirCoupling air) {
        this.assetManager = assetManager;
        this.air = air;
        this.airValve = airValve;

        Material matDebug = new Material(assetManager, "Common/MatDefs/Misc/ShowNormals.j3md");

        Box cube1Mesh = new Box(0.025f, 0.075f, 0.025f);
        GeometryAction airValveGeo = new GeometryAction("AirCoupling", cube1Mesh, this);
        airValveGeo.setMaterial(matDebug);
        airValve.attachChild(airValveGeo);
        airValveGeo.setLocalTranslation(-.075f, .15f, .04f);

        Spatial airLeverMesh = assetManager.loadModel("Models/Obj/AirLever.obj");
        TangentBinormalGenerator.generate(airLeverMesh);
        airLeverMesh.setMaterial(matDebug);
        airValve.attachChild(airLeverMesh);

        close();
    }

    public void open() {
        open = true;
        air.setOpen(open);

        airValve.rotate(-90 * FastMath.DEG_TO_RAD, 0, 0);
    }

    public void close() {
        open = false;
        air.setOpen(open);

        airValve.rotate(90 * FastMath.DEG_TO_RAD, 0, 0);
    }

    @Override
    public void click() {
        open = !open;
        if (open) {
            open();
        } else {
            close();
        }
    }
}
