/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package vehicle;

import a_program.Program;
import a_program.VrPerson;
import com.jme3.asset.AssetManager;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.FastMath;
import com.jme3.math.Quaternion;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Node;
import com.jme3.scene.Spatial;
import com.jme3.scene.shape.Box;
import com.jme3.util.TangentBinormalGenerator;
import e_tco.Action3dElement;
import e_tco.GeometryAction;

/**
 *
 * @author Eirik
 */
public class Coupling implements Action3dElement {

    public enum CouplingState {
        Loose, Held, Connected
    };

    private AssetManager assetManager;
    private Vehicle vehicle;
    private int location; // Where the coupling is located, 1 = front, 2 = back
    private Vector3f target; // Target location for the inverse kinematics
    private Spatial heldBy;
    private CouplingState state = CouplingState.Loose;
    private Coupling connectedTo;

    //  Nodes
    private Node basePivot = new Node();

    private Node link1Pivot = new Node();
    private Node link1Joint1Pivot = new Node();
    private Node link1Joint2Pivot = new Node();

    private Node linkmidPivot = new Node();

    private Node link2Joint1Pivot = new Node();
    private Node link2Joint2Pivot = new Node();
    private Node link3Pivot = new Node();

    private Node linkLeverPivot = new Node();

    private Node endPivot = new Node();
    private Node connectPivot = new Node();

    private Handle handle;

    private InverseKinematics IK = new InverseKinematics(
            new Node[]{basePivot, link1Pivot, link1Joint1Pivot, link3Pivot, endPivot}, true);

    //  Offsets
    private float midOffset = .15f;
    private float minMidOffset = .05f;
    private float maxMidOffset = .175f;
    private float midOffsetSpeed = .001f;
    private float midOffsetDirection = 0;

    private float linkOffset = -.315f;
    private float connectOffset = .1f;
    private float endOffset = .25f;
    private float startToEndLength;

    public Coupling(AssetManager assetManager, Vehicle vehicle, int location) {
        this.assetManager = assetManager;
        this.vehicle = vehicle;
        this.location = location;

        Material matDebug = new Material(assetManager, "Common/MatDefs/Misc/ShowNormals.j3md");
        //mat.setColor("Color", ColorRGBA.Blue);

        Box cube1Mesh = new Box(0.05f, 0.05f, 0.05f);
        GeometryAction kobbelGeo = new GeometryAction("Coupling", cube1Mesh, this);
        kobbelGeo.setMaterial(matDebug);

        TangentBinormalGenerator.setToleranceAngle(179);   // Stops some error warnings, models are funtioning

        basePivot.attachChild(connectPivot);
        connectPivot.setLocalTranslation(0, 0, connectOffset);

        //  Link1
        Spatial link1 = loadModel("Link1");
        link1Pivot.attachChild(link1);
        basePivot.attachChild(link1Pivot);
        //  Joint part 1
        Spatial link1joint1 = loadModel("LinkPivot1");
        link1Joint1Pivot.attachChild(link1joint1);
        link1Pivot.attachChild(link1Joint1Pivot);
        link1Joint1Pivot.setLocalTranslation(0, linkOffset, 0);
        link1Joint1Pivot.rotate(FastMath.DEG_TO_RAD * 90, 0, 0);
        //  Joint part 2
        Spatial link1joint2 = loadModel("LinkPivot2");
        link1Joint2Pivot.attachChild(link1joint2);
        link1Joint1Pivot.attachChild(link1Joint2Pivot);

        //  Middle link
        Spatial linkmid = loadModel("Link2");
        linkmidPivot.attachChild(linkmid);
        link1Joint2Pivot.attachChild(linkmidPivot);
        linkmidPivot.setLocalTranslation(0, 0, midOffset);

        //  Joint 2 part 2
        Spatial link2joint2 = link1joint2.clone();
        link2Joint2Pivot.attachChild(link2joint2);
        linkmidPivot.attachChild(link2Joint2Pivot);
        link2Joint2Pivot.setLocalTranslation(0, 0, midOffset);
        //  Joint 2 part 1
        Spatial link2joint1 = link1joint1.clone();
        link2Joint1Pivot.attachChild(link2joint1);
        link2Joint2Pivot.attachChild(link2Joint1Pivot);
        //  Link3
        Spatial link3 = loadModel("Link3");
        link3Pivot.attachChild(link3);
        link2Joint1Pivot.attachChild(link3Pivot);
        link3Pivot.rotate(FastMath.DEG_TO_RAD * 90, 0, 0);

        //  LinkLever
        Spatial lever = loadModel("LinkLever");
        linkLeverPivot.attachChild(lever);
        linkmidPivot.attachChild(linkLeverPivot);
        linkLeverPivot.setLocalTranslation(0, .04f, 0);
        handle = new Handle(assetManager, linkLeverPivot, this);

        //  LinkEnd
        Spatial end = link1joint1.clone();         //  For debugging
        end.setMaterial(matDebug);                 //  For debugging
        endPivot.attachChild(end);                 //  For debugging
        link3Pivot.attachChild(endPivot);
        endPivot.setLocalTranslation(0, endOffset, 0);

        Spatial connect = end.clone();              //  For debugging
        connect.setMaterial(matDebug);              //  For debugging
        connectPivot.attachChild(connect);          //  For debugging

        endPivot.attachChild(kobbelGeo);            //  For debugging

        target = endPivot.getWorldTranslation();
        startToEndLength = basePivot.getWorldTranslation().distance(target);
    }

    //  Update function, updates every frame
    public void tick(float tpf) {
        switch (state) {
            case Loose:
                target = basePivot.getWorldTranslation().add(new Vector3f(0, -startToEndLength * 1.5f, 0));
                inverseKinematics(target, tpf / 10f);
                break;
            case Held:
                inverseKinematics(target, tpf);
                break;
            case Connected:
                if (connectedTo != null) {
                    target = connectedTo.getConnectPivot().getWorldTranslation();
                    inverseKinematics(target, tpf);
                }
                break;
        }

        rotateHandle(midOffsetDirection);
    }

    public Node getBasePivot() {
        return basePivot;
    }

    public Node getEndPivot() {
        return endPivot;
    }

    public Node getConnectPivot() {
        return connectPivot;
    }

    public Coupling getConnectedTo() {
        return connectedTo;
    }

    public Vehicle getVehicle() {
        return vehicle;
    }

    public void setConnectedTo(Coupling c) {
        connectedTo = c;
    }

    public void connectTo(Coupling coupling) {
        state = CouplingState.Connected;
        connectedTo = coupling;
        coupling.setConnectedTo(this);
        heldBy = null;
        Program.updateCouplings();
    }

    public void disconnect() {
        state = CouplingState.Loose;
        if (connectedTo.getState() != CouplingState.Connected) {
            connectedTo.setConnectedTo(null);
            connectedTo = null;
        }
        Program.updateCouplings();
    }

    public void rotateHandle(float r) {
        if (state == CouplingState.Connected) {
            float length = basePivot.getWorldTranslation().distance(link1Pivot.getWorldTranslation());
            length += link1Pivot.getWorldTranslation().distance(link1Joint1Pivot.getWorldTranslation());
            length += link1Joint1Pivot.getWorldTranslation().distance(link3Pivot.getWorldTranslation());
            length += link3Pivot.getWorldTranslation().distance(endPivot.getWorldTranslation());
            float connectionLength = basePivot.getWorldTranslation().distance(connectedTo.getConnectPivot().getWorldTranslation());
            //System.out.println("l:" + length + ", cl:" + connectionLength + ", r:" + r*midOffsetSpeed);

            if (r < 0 && length < connectionLength) {
                return;
            }
        }

        float prev = midOffset;
        midOffset += midOffsetSpeed * r;
        midOffset = FastMath.clamp(midOffset, minMidOffset, maxMidOffset);
        linkmidPivot.setLocalTranslation(0, 0, midOffset);
        link2Joint2Pivot.setLocalTranslation(0, 0, midOffset);
        linkmidPivot.rotate(0, 0, FastMath.DEG_TO_RAD * (midOffset - prev) * 10000);
        link2Joint2Pivot.rotate(0, 0, FastMath.DEG_TO_RAD * -(midOffset - prev) * 10000);
    }

    public void interactMidOffset() {
        if (midOffsetDirection == -1) {
            midOffsetDirection = 1;
        } else {
            midOffsetDirection = -1;
        }
    }

    public Vector3f getHandlePos() {
        return handle.getHandle();
    }

    public void setTarget(Vector3f target) {
        this.target = target;
    }

    public void hold(Spatial heldBy) {
        setState(CouplingState.Held);
        this.heldBy = heldBy;
        setTarget(heldBy.getWorldTranslation());
    }

    public void release() {
        heldBy = null;
        setState(CouplingState.Loose);
    }

    public Spatial heldBy() {
        return heldBy;
    }

    public CouplingState getState() {
        return state;
    }

    public void setState(CouplingState s) {
        state = s;
    }

    public void inverseKinematics(Vector3f target, float tpf) {
        IK.setActive(true);
        IK.setTarget(target);
        IK.controlUpdate(tpf);
    }

    private Spatial loadModel(String name) {
        Spatial mesh = assetManager.loadModel("Models/Obj/" + name + ".obj");
        TangentBinormalGenerator.generate(mesh);
        Material mat = loadMaterial(name);
        mesh.setMaterial(mat);

        return mesh;
    }

    private Material loadMaterial(String name) {
        Material mat = new Material(assetManager, "Common/MatDefs/Light/Lighting.j3md");
        mat.setTexture("DiffuseMap", assetManager.loadTexture("Textures/Train/Link_m_" + name + "_BaseColor.png"));
        mat.setTexture("NormalMap", assetManager.loadTexture("Textures/Train/Link_m_" + name + "_Normal.png"));
        mat.setBoolean("UseMaterialColors", true);
        mat.setColor("Diffuse", ColorRGBA.White);  // minimum material color
        mat.setColor("Specular", ColorRGBA.White); // for shininess
        mat.setFloat("Shininess", 64f); // [1,128] for shininess

        return mat;
    }

    private void attemptConnect() {
        VrPerson.connect(this);
    }

    @Override
    public void click() {
        if (state != CouplingState.Connected) {
            attemptConnect();
        } else {
            disconnect();
        }
    }
}

//------------------------------------
//  Handle
//------------------------------------
class Handle implements Action3dElement {

    AssetManager assetManager;
    Coupling coupling;
    Node handle;
    GeometryAction handleGeo;

    public Handle(AssetManager assetManager, Node handle, Coupling coupling) {
        this.assetManager = assetManager;
        this.coupling = coupling;
        this.handle = handle;

        Material matDebug = new Material(assetManager, "Common/MatDefs/Misc/ShowNormals.j3md");

        Box cube1Mesh = new Box(0.025f, 0.025f, 0.025f);
        handleGeo = new GeometryAction("CouplingHandle", cube1Mesh, this);
        handleGeo.setMaterial(matDebug);
        handle.attachChild(handleGeo);
        handleGeo.setLocalTranslation(0, .2f, 0);
    }

    public Vector3f getHandle() {
        return handleGeo.getWorldTranslation();
    }

    @Override
    public void click() {
        coupling.interactMidOffset();
    }
}
